/*
 *
 * Copyright (C) 2002-2003 Rich Moore <rich@kde.org>
 * Copyright (C) 2002-2003 George Staikos <staikos@kde.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <qdir.h>
#include <qpushbutton.h>
#include <qsplitter.h>
#include <qtimer.h>
#include <qtextedit.h>

#include <kaction.h>
#include <kconfig.h>
#include <kdebug.h>
#include <kkeydialog.h>
#include <klocale.h>
#include <kmenubar.h>
#include <kpopupmenu.h>
#include <krun.h>
#include <kstandarddirs.h>
#include <kstatusbar.h>
#include <ktoolbar.h>
#include <kapplication.h>

#include <kwin.h>
#include <netwm.h>

#include "qtvisionwidget.h"
#include "viewmanager.h"
#include "qtvision.h"
#include "qtvisionactions.h"
#include "actions.h"
#include "sidebar.h"
#include "qtvisionsystray.h"
#include "savercontrol.h"
#include "dpmscontrol.h"
#include "osdmanager.h"
#include "cfgdata.h"
#include "changelogdialog.h"
#include "checkupdatesdialog.h"
#include "statusmanager.h"


#define STATUSBAR_MAIN		0
#define STATUSBAR_PROGRAM	1

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#ifdef HAVE_LIRC_QTVISION
#include "lirc.h"
#endif // HAVE_LIRC_QTVISION

#include "mainwindow.h"
#include "mainwindow.moc"

MainWindow::MainWindow( QWidget *parent, const char *name, WFlags f )
    : DCOPObject( "QtVisionViewIface" ),
      KMainWindow( parent, name ? name : "main_window", f ), 
      _systray(0), _lirc(0), _viewMode(ViewNormal), _sleepdelay(0), _savedGuiStateValid(false)
{
    _toolbarVisibilityList.setAutoDelete(true);

    // Views
    _view = new QSplitter( this );
    _side = new Sidebar( _view );
    _screen = new QtVisionWidget( _view );

    setCentralWidget( _view );

    _cfg = _screen->driver()->configHandle();

    // Create the sidebar
    _side->setScreen( _screen );
    _side->setFocusProxy(_screen);
    _side->createViews();

    // Create the actions
    actionCollection()->setHighlightingEnabled( true );
    createActions();

    // Setup status messages
    connect( actionCollection(), SIGNAL( actionStatusText(const QString &) ),
             this, SLOT( statusMessage(const QString & ) ) );
    connect( actionCollection(), SIGNAL( clearStatusText() ),
             this, SLOT( clearStatusBar() ) );
#if 0    
    connect( _screen->driver(), SIGNAL( channelChanged(const QString &) ),
             statusBar(), SLOT( permanent(const QString & ) ) );
#endif
    connect( _screen->driver(), SIGNAL( channelChanged(const QString &) ),
             this, SLOT( setCaption(const QString & ) ) );  
    connect( _screen->driver()->views(), SIGNAL( stayOnTopChanged(bool) ),
             this, SLOT( stayOnTopCfgChanged(bool) ) );
    connect( _screen->driver()->views(), SIGNAL( disableScreenSaver(bool, int) ),
             this, SLOT( screenSaverCfgChanged(bool, int) ) );
    connect( _screen->view(), SIGNAL( doubleClicked() ),
             this, SLOT( nextViewMode() ) );
    connect( this, SIGNAL( numberKeyPressed(int) ),
             _screen->driver(), SLOT( processNumberKeyEvent(int) ) );

    // Enable/disable actions
    _actions->deviceChanged( 0 );

    _sleeptimer = new QTimer(parent);
    connect(_sleeptimer, SIGNAL(timeout()), this, SLOT(slotSleepTimerTimeout()));

    connect(this, SIGNAL(sleepTimerDelay(const QString &)),
	_screen->driver()->osdManager(), SLOT(displayCC(const QString &)));
    
    // connect actions to widgets
    connect(_actShowSidebar, SIGNAL( toggled(bool) ),
	    _side,           SLOT( toggleView(bool) ));
    connect(_side,           SIGNAL( visibilityChanged(bool) ),
	    _actShowSidebar, SLOT( setChecked(bool) ));

    setupInfraRed();

    // Create the main window shell
    setXMLFile( "qtvisionui.rc" );
    createGUI();

    KMenuBar *mb = menuBar();
    mb->changeItem( mb->idAt(0), i18n("&Channels") );

    // setup import entries for channel files
    QPopupMenu* iMenu = static_cast<QPopupMenu*>(guiFactory()->container("import", this));
    if (iMenu) {
	const QStringList& s = _screen->driver()->fileFormatsRead();
	for (QStringList::const_iterator it = s.begin();
	     it != s.end();
	     ++it) {
	    ImpExChannelsAction* a;
	    a= new ImpExChannelsAction(*it + i18n(" file..."), 0, *it,
				       _screen->driver(), 1, actionCollection(), *it);
	    a->plug(iMenu);
	}
    } else {
	kdWarning() << "Import menu GUI element not found, please delete you qtvisionui.rc file in your .kde dir (you will loose your configured keyboard shortcuts)" << endl;
    }

    // setup export entries for channel files
    QPopupMenu* eMenu = static_cast<QPopupMenu*>(guiFactory()->container("export", this));
    if (eMenu) {
	const QStringList& s = _screen->driver()->fileFormatsWrite();
	for (QStringList::const_iterator it = s.begin();
	     it != s.end();
	     ++it) {
	    ImpExChannelsAction* a;
	    a = new ImpExChannelsAction(*it + i18n(" file..."), 0, *it,
					_screen->driver(), 0, actionCollection(), *it);
	    a->plug(eMenu);
	}
    } else {
	kdWarning() << "Export menu GUI element not found, please delete you qtvisionui.rc file in your .kde dir (you will loose your configured keyboard shortcuts)" << endl;
    }

    // Create context menu
    insertContextMenu( "screen_context_popup" );

    // Store the current status of the screensaver
    _screenSaverWasEnabled = SaverControl::getSaverStatus();
    // and the dpms
    _dpmsWasEnabled = DpmsControl::getDpmsStatus();

    // Check the screensaver configuration & act accordingly
    screenSaverCfgChanged( _screen->driver()->config()->disableScreenSaver,
                           _screen->driver()->config()->screenSaverMode );

    connect( this, SIGNAL( visibilityChanged(bool) ),
	     this, SLOT( disableScreenSaver(bool) ) );

    // set stayOnTop according to the config
    stayOnTopCfgChanged( _screen->driver()->config()->stayOnTop );

    // Create the status bar
    statusBar()->insertItem("", STATUSBAR_MAIN, 1);
    statusBar()->setItemAlignment(STATUSBAR_MAIN, AlignLeft | AlignVCenter);
    statusBar()->insertItem("", STATUSBAR_PROGRAM, 0, true);

    connect(_screen->driver()->statusManager(),
		SIGNAL(messageSent(const QString&)),
		this, SLOT(shortMessage(const QString&)));

    connect(_screen->driver()->statusManager(),
		SIGNAL(programChanged(const QString&)),
		this, SLOT(sbProgramName(const QString&)));

    // allow installation of plugin gui elements
    _screen->driver()->setGuiFactory(guiFactory(), actionCollection());

    // Apply main window settings & size
    restoreWindowSettings();
    _size = size();
    _pos  = pos();

    QTimer::singleShot( 10, _screen->driver(), SLOT(start()) );
    QTimer::singleShot( 10, this, SLOT(restoreViewMode()) );
}

MainWindow::~MainWindow()
{
    // If the screensaver was previously enabled, re-enable it.
    // May cause confusion if the screensaver's configuration
    // has been changed while this application was running...
    disableScreenSaver( false );
    
    // Save the current window settings & size    
    saveWindowSettings();

    delete _sleeptimer;

    // So we don't crash on exit
    delete _actions;    
}

void MainWindow::createActions()
{
    KActionCollection* acts = actionCollection();

    // Standard actions
    (void) KStdAction::quit( qApp, SLOT( quit() ), acts );
    _actShowMenubar   = KStdAction::showMenubar( this, SLOT( toggleMenubar() ), acts );
    _actShowStatusbar = KStdAction::showStatusbar( this, SLOT( toggleStatus() ), acts );
    _actShowToolbar   = KStdAction::showToolbar("mainToolBar", acts);

    // QtVision actions
    _actions = new QtVisionActions( _screen->driver() );
    _actions->createActions( acts );

    connect( _screen->driver(), SIGNAL( channelText(const QString &) ),
             _actions->channel(), SLOT( display(const QString &) ) );

    connect( _actions, SIGNAL(launchBindings()), this, SLOT(keyBindings()) );

    // View Modes
    _actShowNorm = new KRadioAction( i18n("&Normal"),        "kwintv",            Key_Home,     acts, "show_normal" );
    _actShowTop  = new KRadioAction( i18n("T&V Mode"),       "inline_image",      Key_PageDown, acts, "show_top_level" );
    _actShowFull = new KRadioAction( i18n("&Full Screen"),   "window_fullscreen", Key_PageUp,   acts, "full_screen" );
    _actShowDesk = new KRadioAction( i18n("Video &Desktop"), "frame_image",       0,            acts, "show_video_desktop" );
    _actShowNorm->setExclusiveGroup( "view_modes" );
    _actShowTop->setExclusiveGroup(  "view_modes" );
    _actShowFull->setExclusiveGroup( "view_modes" );
    _actShowDesk->setExclusiveGroup( "view_modes" );
    connect( _actShowNorm, SIGNAL(toggled(bool)), this, SLOT(actSetViewModeNormal(bool)) );
    connect( _actShowTop,  SIGNAL(toggled(bool)), this, SLOT(actSetViewModeTopLevel(bool)) );
    connect( _actShowFull, SIGNAL(toggled(bool)), this, SLOT(actSetViewModeFullScreen(bool)) );
    connect( _actShowDesk, SIGNAL(toggled(bool)), this, SLOT(actSetViewModeVideoDesktop(bool)) );

    _actNextView = new KAction( i18n("Next View"), "next", 0, this, SLOT( nextViewMode() ), acts,
				"next_view_mode" );
    _actToggleFull = new KToggleAction( i18n("Toggle FullScreen"), "window_fullscreen", Key_F,
					this, SLOT( toggleFullScreen() ),
					acts, "toggle_fullscreen" );

    // GUI elements
    _actShowSidebar = new KToggleAction( i18n("Show Sidebar"), "view_sidetree", Key_F9,
					 acts, "show_sidebar" );
    _actShowSidebar->setStatusText( i18n( "Shows and hides the sidebar." ) );

    _actShowViewbar = new KToggleToolBarAction( "viewToolBar", i18n("Show View Toolbar"),
						acts, "show_view_toolbar" );
    _actShowViewbar->setIcon("wizard");

    _actShowVolbar = new KToggleToolBarAction( "volumeToolBar", i18n("Show Volume Toolbar"),
					       acts, "show_volume_toolbar" );
    _actShowVolbar->setIcon("ktv_volup");

    // Help actions
    _actChangeLog = new KAction( i18n("&Change Log..."), "log", 0,
                                  this, SLOT( showChangeLog() ),
                                  acts, "help_view_changelog" );

    _actCVSLog = new KAction( i18n("C&VS Change Log..."), "html", 0,
                               this, SLOT( showCVSLog() ),
                               acts, "help_view_cvslog" );
    _actWebSite = new KAction( i18n("KWinTV Web Site..."), "html", 0,
                                this, SLOT( showWebSite() ),
                                acts, "help_view_website" );

    _sleepTime = new KToggleAction( i18n("Sleep Timer"), "ktv_sleeptimer", Key_T,
                                 this, SLOT( sleepTimer() ), acts, 
                                 "sleep_timer" );
}

void MainWindow::setupInfraRed()
{
#ifdef HAVE_LIRC_QTVISION
    
    // Remove the instance in case this function gets invoked
    // twice.  This should not happen.  However, since we can
    // not be certain about that it is always better to be safe
    // than sorry...
    if (_lirc != 0)
      delete _lirc;
    
    // Attach it to this window, not its parent.  Actually
    // parent () in this case will return 0L since this is
    // the top-level window...
    _lirc = new Lirc (this, "qtvision");
    
    QMap<QString, QString> lircKeyMap;
    lircKeyMap["CH+"] = "channel_up";
    lircKeyMap["CH-"] = "channel_down";
    lircKeyMap["VOL+"] = "volume_up";
    lircKeyMap["VOL-"] = "volume_down";
    lircKeyMap["CH_UP"] = "channel_up";
    lircKeyMap["CH_DOWN"] = "channel_down";
    lircKeyMap["VOL_UP"] = "volume_up";
    lircKeyMap["VOL_DOWN"] = "volume_down";
    lircKeyMap["MUTE"] = "volume_mute";
    lircKeyMap["QUIT"] = "file_quit";
    lircKeyMap["POWER"] = "file_quit";
    //(*lircKeyMap)["FULL_SCREEN"] = "next_view_mode";
    //(*lircKeyMap)["FULLSCREEN"] = "next_view_mode";
    lircKeyMap["FULL_SCREEN"] = "toggle_fullscreen";
    lircKeyMap["FULLSCREEN"] = "toggle_fullscreen";
    lircKeyMap["TTX"] = "toggle_teletext";
    lircKeyMap["CC"] = "toggle_teletext";
    
    _lirc->setDefaultKeyMap(lircKeyMap);
    connect(_lirc, SIGNAL(command(const QString&, unsigned)),
                  SLOT(lircCommand(const QString&, unsigned)));
    connect(_lirc, SIGNAL(event(const QString&, unsigned)),
                  SLOT(lircCommand(const QString&, unsigned)));
#else
    kdWarning() << "MainWindow::setupInfraRed(): Lirc not available" << endl;
#endif // HAVE_LIRC_QTVISION
}

KPopupMenu *MainWindow::createMenu( const QString &name ) 
{
    KPopupMenu *pop = static_cast<KPopupMenu *>( factory()->container(name, this) );
    if ( !pop )
        return 0;

    pop->insertTitle( i18n("QtVision"), -1, 0 );
    return pop;
}

KPopupMenu *MainWindow::insertContextMenu( const QString &name ) 
{
    KPopupMenu *pop = createMenu( name );
    _screen->addContextMenu( pop );

    return pop;
}

void MainWindow::toggleStatus()
{
    if ( statusBar()->isVisible())
        statusBar()->hide();
    else
        statusBar()->show();
}

void MainWindow::toggleMenubar()
{
    if ( menuBar()->isVisible() )
        menuBar()->hide();
    else
        menuBar()->show();
}

void MainWindow::setShowTrayIcon( bool show )
{
    if ( show ) {
        if ( !_systray ) {
            kdDebug() << "Creating systray icon" << endl;
            _systray = new QtVisionSysTray(this, "qtvision_systray",
					   createMenu("system_tray_popup"), _actions);
            connect(_systray, SIGNAL(doubleClicked()), this, SLOT(setViewModeNormal()));
            connect(_systray, SIGNAL(leftClicked()), this, SLOT(bringToFront()));
        }
        kdDebug() << "Showing systray icon" << endl;
        _systray->show();
    }
    else {
        if ( _systray ) {
            _systray->hide();
        }
    }
}

void MainWindow::disableScreenSaver( bool disable )
{
    if ( disable ) {
        SaverControl::disableSaver();
        DpmsControl::disableDpms();
    }
    else {
        if ( _screenSaverWasEnabled )
            SaverControl::enableSaver();
        
        if ( _dpmsWasEnabled )
            DpmsControl::enableDpms();
    }
}

void MainWindow::screenSaverCfgChanged( bool disable, int mode ) {
    _activateVisibility = false;
    if ( disable ) {
	screenSaverModeChanged( mode );
    } else {
	if ( _screenSaverWasEnabled ) 
	    SaverControl::enableSaver();
    }
}

void MainWindow::screenSaverModeChanged( int mode )
{
    switch( mode ) {
    case SAVER_RUNNING:
        disableScreenSaver( true );
        break;
    case SAVER_VISIBLE:
        _activateVisibility = true;
        checkScreenSaver();
        break;
    case SAVER_FULLSCREEN:
        if ( _viewMode == ViewFullScreen ) {
            disableScreenSaver( true );
        } else {
            disableScreenSaver( false );
        }
        break;
    default:
        kdDebug() << "MainWindow: We should never get here!" << endl;
        break;
    }    
}

void MainWindow::saveWindowSettings()
{
    _cfg->setGroup( "Window Geometry" );

    _cfg->writeEntry( "View Mode", _viewMode );    
    setViewModeNormal();

    _cfg->writeEntry( "Window Position X", x() );
    _cfg->writeEntry( "Window Position Y", y() );
    _cfg->writeEntry( "Show Sidebar", !_side->isHidden() );
    _cfg->writeEntry( "Splitter Position", _view->sizes() );    
    _cfg->writeEntry( "TopLevel Pos", _topLevelPos );
    _cfg->writeEntry( "TopLevel Size", _topLevelSize );

    saveMainWindowSettings(_cfg, "MainWindow");
}

void MainWindow::restoreWindowSettings()
{
    applyMainWindowSettings(_cfg, "MainWindow");

    // update slider orientation
    _actions->volumeSlider()->dockChange(toolBar("volumeToolBar"));

    // these are just toggleActions (they do not track widget state)
    _actShowMenubar->setChecked(!menuBar()->isHidden());
    _actShowStatusbar->setChecked(!statusBar()->isHidden());

    _cfg->setGroup( "Window Geometry" );
    int x = _cfg->readNumEntry( "Window Position X", 0 );
    int y = _cfg->readNumEntry( "Window Position Y", 0 );
    move( x, y );

    QSize defS(400, 300);
    _topLevelPos  = _cfg->readPointEntry("TopLevel Pos");
    _topLevelSize = _cfg->readSizeEntry("TopLevel Size", &defS);

    QValueList<int> sizes = _cfg->readIntListEntry( "Splitter Position" );
    if ( !sizes.count() ) {
        sizes.append( 20 );
        sizes.append( 80 );
    }
    _view->setSizes( sizes );

    bool sidebar = _cfg->readBoolEntry( "Show Sidebar", true );
    if ( sidebar )
        _side->show();
    else
        _side->hide();
}

void MainWindow::restoreViewMode()
{
    _cfg->setGroup( "Window Geometry" );
    setViewMode( _cfg->readNumEntry("View Mode", ViewNormal) );
}

void MainWindow::showChangeLog()
{
    ChangeLogDialog *dlg = new ChangeLogDialog( this, "changelog", false, 
                                                Qt::WType_Dialog | 
                                                Qt::WDestructiveClose );
    connect( dlg->closeButton, SIGNAL( clicked() ), dlg, SLOT( close() ) );

    QString log = locate("appdata", "ChangeLog");
    QFile file( log.isNull() ? QString("ChangeLog") : log );

    if ( file.open( IO_ReadOnly ) ) {
        QByteArray data = file.readAll();
        QString s( data );
        QFont f = dlg->edit->font();
        f.setFamily( "courier" );
        dlg->edit->setFont( f );
        dlg->edit->setText( s );
    }

    dlg->show();
}

void MainWindow::showCVSLog()
{
    KURL url( i18n("http://www.kwintv.org/rewrite_changelog.html") );
    KRun::runURL( url, QString("text/html") );
}

void MainWindow::showWebSite()
{
    KURL url( i18n("http://www.kwintv.org/") );
    KRun::runURL( url, QString("text/html") );
}

void MainWindow::lircCommand(const QString &cmd, unsigned /*repeat*/)
{
    if (cmd.length() == 1 && cmd[0] >= '0' && cmd[0] <= '9') {
        // handle numpad keys
        int key = cmd.toInt();
        emit numberKeyPressed(key);
    }

    KAction* act = action( cmd.latin1() );
    if (act)
      act->activate();
    else
        kdDebug() << "lirc: command '" << cmd << "' not recognized" << endl;
}

void MainWindow::keyBindings()
{
    KKeyDialog::configure(actionCollection());
}

void MainWindow::stayOnTopCfgChanged(bool enable)
{
    _staysOnTop = enable;
    checkStayOnTop();
}

void MainWindow::checkStayOnTop()
{
    // Dirk: FIXME: Doesn't work reliably for me.
    if (_staysOnTop) {
	setWFlags(getWFlags() | WStyle_Customize | WStyle_StaysOnTop);
    } else {
	setWFlags(getWFlags() & ~WStyle_StaysOnTop );
    }
}

void MainWindow::sbProgramName(const QString& name) {
    statusBar()->changeItem(name, STATUSBAR_PROGRAM);
}

void MainWindow::statusMessage(const QString& name) {
    statusBar()->changeItem(name, STATUSBAR_MAIN);
}

void MainWindow::shortMessage(const QString& name) {
    statusBar()->changeItem(name, STATUSBAR_MAIN);
    QTimer::singleShot(2000, this, SLOT(clearStatusBar()));
}

void MainWindow::clearStatusBar() {
    statusBar()->changeItem("", STATUSBAR_MAIN);
}

void MainWindow::slotSleepTimerTimeout()
{
    kdDebug() << "QtVision: sleepTimer timeout reached - quitting client." << endl;
    QApplication::exit();
}

void MainWindow::sleepTimer()
{
    if (_sleeptimer->isActive())
        _sleeptimer->stop();

    kdDebug() << "sleepTimer() called.." << endl;

    if (_sleepdelay == 0) {
        // Change to 15 minutes.
        _sleepdelay = 900000;

    } else if (_sleepdelay == 900000) {
        // Change to 30 minutes.
        _sleepdelay = 1800000;

    } else if (_sleepdelay == 1800000) {
        // Change to 60 minutes.
        _sleepdelay = 3600000;

    } else if (_sleepdelay == 3600000) {
        // Change to 120 minutes.
        _sleepdelay = 7200000;

    } else {
        // Remove sleep-delay
        _sleepdelay = 0;
    }

    if (_sleepdelay)
          _sleeptimer->start(_sleepdelay, true);  // start the count-down timer

    // Display to the screen, how many minutes there are to the TV turns
    // off.
    emit sleepTimerDelay(QString::number(_sleepdelay / 1000 / 60));
}

void MainWindow::nextViewMode()
{
    kdDebug() << "MainWindow::nextViewMode() Current is " << _viewMode << endl;

    // Note: We don't want to switch to the ViewNone mode
    int viewMode = _viewMode;
    if ( ++viewMode > ViewMaxIterable )
	viewMode = ViewNormal;
    setViewMode( viewMode );
}

void MainWindow::toggleFullScreen()
{
    if ( _viewMode == ViewFullScreen ) {
	setViewMode( (_oldViewMode <= ViewMaxIterable) ? _oldViewMode : ViewNormal );
    } else {
	_oldViewMode = _viewMode;
	setViewMode( ViewFullScreen );
    }
}

void MainWindow::setViewMode( int mode )
{
    _changingViewMode = true;
    doSetViewMode( _viewMode, false );
    checkStayOnTop();
    doSetViewMode( mode,      true  );
    _changingViewMode = false;
    emit viewModeChanged( _viewMode );
}

void MainWindow::doSetViewMode( int mode, bool enable )
{
    switch( mode ) {
    case ViewNone:
	setShowNone( enable );
	break;
    case ViewNormal:
	setShowNormal( enable );
	break;
    case ViewTopLevel:
	setShowTopLevel( enable );
	break;
    case ViewFullScreen:
	setShowFullScreen( enable );
	break;
    case ViewVideoDesktop:
	setShowVideoDesktop( enable );
	break;
    default:
	kdDebug() << "MainWindow::setViewMode(): Unknown mode " << mode
		  << ", defaulting to ViewNormal" << endl;
	setShowNormal( enable );
	break;
    }
}

void MainWindow::setShowNone( bool enable )
{
    kdDebug() << "MainWindow::setShowNone(): " << enable << endl;
    if (enable) {
	_screen->driver()->stop();
	hide();
	_viewMode = ViewNone;
    } else {
	show();
	_screen->driver()->start();
    }
}

void MainWindow::setShowNormal( bool enable )
{
    kdDebug() << "MainWindow::setShowNormal(): " << enable << endl;
    _viewMode = ViewNormal;
    _actShowNorm->setChecked( true );
}

void MainWindow::setShowTopLevel( bool enable )
{
    kdDebug() << "MainWindow::setShowTopLevel(): " << enable << endl;
    if (enable) {
	hideGUIElements();
	setShowTrayIcon( true );
	setWFlags( WStyle_Customize | WStyle_Tool );
	move( _topLevelPos );
	resize( _topLevelSize );
	QTimer::singleShot(1, this, SLOT(bringToFront()));
	_viewMode = ViewTopLevel;
	_actShowTop->setChecked( true );
    } else {
	clearWFlags( WStyle_Customize | WStyle_Tool );
	setShowTrayIcon( false );
	restoreGUIElements();
	move( _pos );
	resize( _size );
    }
}

void MainWindow::setShowFullScreen( bool enable )
{
    kdDebug() << "MainWindow::setShowFullScreen(): " << enable << endl;
    if (enable) {
	// icons are lost at showFullScreen()
	_miniIcon = KWin::icon( winId(), 16, 16, true );
	_icon     = KWin::icon( winId(), 34, 34, false );
	if ( _icon.isNull() ) {
	    _icon = KWin::icon( winId(), 32, 32, true );
	}
	hideGUIElements();
	setWFlags( WStyle_Customize | WStyle_NoBorder | WStyle_StaysOnTop );
	showFullScreen();
	QTimer::singleShot(1, this, SLOT(bringToFront()));
	_viewMode = ViewFullScreen;
	_actShowFull->setChecked( true );
    } else {
	clearWFlags( WStyle_Customize | WStyle_NoBorder | WStyle_StaysOnTop );
	restoreGUIElements();
	showNormal();
	KWin::setIcons( winId(), _icon, _miniIcon );
    }
}

void MainWindow::setShowVideoDesktop( bool enable )
{
    kdDebug() << "MainWindow::setShowVideoDesktop(): " << enable << endl;
    if (enable) {
	_screen->driver()->setVideoDesktop( true );
	setShowTrayIcon( true );
	hide();
	_viewMode = ViewVideoDesktop;
	_actShowDesk->setChecked( true );
    } else {
	show();
	setShowTrayIcon( false );
	_screen->driver()->setVideoDesktop(false);
    }
}

void MainWindow::checkScreenSaver()
{
    if ( _activateVisibility ) {
	kdDebug() << "disable saver: " << !isHidden() << endl;
	disableScreenSaver( !isHidden() );
    }
}

void MainWindow::bringToFront()
{
    setActiveWindow();
    raise();
}

void MainWindow::hideGUIElements()
{
    _menubarVisible = !menuBar()->isHidden();
    menuBar()->hide();
    _actShowMenubar->setEnabled(false);

    _statusbarVisible = !statusBar()->isHidden();
    statusBar()->hide();
    _actShowStatusbar->setEnabled(false);

    _sidebarVisible = !_side->isHidden();
    _side->hide();
    _actShowSidebar->setEnabled(false);

    _toolbarVisibilityList.clear();
    KToolBar *toolbar = 0;
    QPtrListIterator<KToolBar> it( toolBarIterator() );
    while ( ( toolbar = it.current() ) != 0 ) {
        ++it;
	_toolbarVisibilityList.append(new toolBarVisibilityPair(toolbar, !toolbar->isHidden()));
	toolbar->hide();
    }

    _savedGuiStateValid = true;
}

void MainWindow::restoreGUIElements()
{
    if (!_savedGuiStateValid) return;

    if (_menubarVisible) {
	menuBar()->show();
	_actShowMenubar->setChecked(true);
    }
    _actShowMenubar->setEnabled(true);

    if (_statusbarVisible) {
	statusBar()->show();
	_actShowStatusbar->setChecked(true);
    }
    _actShowStatusbar->setEnabled(true);

    if (_sidebarVisible) {
	_side->show();
    }
    _actShowSidebar->setEnabled(true);

    for(toolBarVisibilityPair *it = _toolbarVisibilityList.first();
	it;
	it = _toolbarVisibilityList.next()) {
	if (it->second) {
	    it->first->show();
	}
    }
}

void MainWindow::moveEvent( QMoveEvent* e )
{
    if (_changingViewMode) return;

    if (_viewMode == ViewTopLevel) {
        _topLevelPos = e->pos();
    } else if (_viewMode == ViewNormal) {
	_pos = e->pos();
    }
}

void MainWindow::resizeEvent( QResizeEvent* e )
{
    if (_changingViewMode) return;

    if (_viewMode == ViewTopLevel) {
	_topLevelSize = e->size();
    } else if (_viewMode == ViewNormal) {
	_size = e->size();
    }
}

void MainWindow::hideEvent( QEvent* )
{
    emit visibilityChanged(false);
}

void MainWindow::showEvent( QEvent* )
{
    emit visibilityChanged(true);
}
